# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


MULTIPLE_SENSITIVE_FEATURE_COMPRESSION_SKIP_REASON = \
    "Multiple sensitive features cannot be compressed into one-dimensional data structure."
