# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

SCORE_KEY = "score"
LABEL_KEY = "label"
SENSITIVE_FEATURE_KEY = "sensitive_feature"
P0_KEY = "p0"
P1_KEY = "p1"

OUTPUT_SEPARATOR = "-"*65

DEMOGRAPHIC_PARITY = "demographic_parity"
EQUALIZED_ODDS = "equalized_odds"

_MATPLOTLIB_IMPORT_ERROR_MESSAGE = "Please make sure to install fairlearn[customplots] to use " \
                                   "the postprocessing plots."
